/**
 * ==============================================
 *  DQL : Descartes Query Language
 * ==============================================
 *
 * (c) Copyright 2014-2016, by Juergen Walter, Fabian Gorsler, Fabian Brosig and other contributors.
 *
 * Project Info:   http://descartes.tools/dql
 *
 * All rights reserved. This software is made available under the terms of the
 * Eclipse Public License (EPL) v1.0 as published by the Eclipse Foundation
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This software is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the Eclipse Public License (EPL)
 * for more details.
 *
 * You should have received a copy of the Eclipse Public License (EPL)
 * along with this software; if not visit http://www.eclipse.org or write to
 * Eclipse Foundation, Inc., 308 SW First Avenue, Suite 110, Portland, 97204 USA
 * Email: license (at) eclipse.org
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package tools.descartes.dql.core.engine.whatif;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceReference;

import tools.descartes.dql.connector.ConnectorProvider;
import tools.descartes.dql.connector.ModelStructureQueryConnector;
import tools.descartes.dql.connector.PerformanceMetricsQueryConnector;
import tools.descartes.dql.connector.QueryConnector;
import tools.descartes.dql.core.engine.Activator;
import tools.descartes.dql.core.engine.query.QueryExecutionEngine;
import tools.descartes.dql.core.engine.query.QueryExecutionEngineImpl;
import tools.descartes.dql.core.engine.util.DQLLogger;
import tools.descartes.dql.core.registry.ConnectorRegistry;
import tools.descartes.dql.lang.descartesQL.AliasClause;
import tools.descartes.dql.lang.descartesQL.ArithmeticClause;
import tools.descartes.dql.lang.descartesQL.DescartesQL;
import tools.descartes.dql.lang.descartesQL.DescartesQLFactory;
import tools.descartes.dql.lang.descartesQL.EntityReference;
import tools.descartes.dql.lang.descartesQL.ForClause;
import tools.descartes.dql.lang.descartesQL.IntervalVariationClause;
import tools.descartes.dql.lang.descartesQL.ListMetricsQuery;
import tools.descartes.dql.lang.descartesQL.ListQuery;
import tools.descartes.dql.lang.descartesQL.MetricClause;
import tools.descartes.dql.lang.descartesQL.MetricClauseType;
import tools.descartes.dql.lang.descartesQL.MetricReference;
import tools.descartes.dql.lang.descartesQL.MetricReferenceClauses;
import tools.descartes.dql.lang.descartesQL.UsingClause;
import tools.descartes.dql.lang.descartesQL.ValueVariationClause;
import tools.descartes.dql.lang.descartesQL.VariationInstruction;
import tools.descartes.dql.lang.descartesQL.WhatIfQuery;
import tools.descartes.dql.lang.descartesQL.WhatIfVariationClause;
import tools.descartes.dql.models.mapping.domain.DecimalResult;
import tools.descartes.dql.models.mapping.domain.DomainFactory;
import tools.descartes.dql.models.mapping.domain.Entity;
import tools.descartes.dql.models.mapping.domain.Probe;
import tools.descartes.dql.models.mapping.domain.Result;
import tools.descartes.dql.models.mapping.domain.StatType;
import tools.descartes.dql.models.mapping.mapping.EntityMapping;
import tools.descartes.dql.models.mapping.mapping.MappingFactory;

public class WhatIfUtil {

	private static final Logger logger = DQLLogger.getLogger(WhatIfQueryInterpreter.class.getName());
	private static final DomainFactory domainFactory = DomainFactory.eINSTANCE;
	private static final MappingFactory mappingFactory = MappingFactory.eINSTANCE;
	private final static DescartesQLFactory dqlFactory = DescartesQLFactory.eINSTANCE;

	public static PerformanceMetricsQueryConnector getPmqc(String family) {
		ConnectorProvider provider = getConnectorProvider(family, PerformanceMetricsQueryConnector.class);
		return provider.createPerformanceMetricsQueryConnectorInstance();
	}

	public static ModelStructureQueryConnector getMsqc(String family) {
		ConnectorProvider provider = getConnectorProvider(family, ModelStructureQueryConnector.class);
		return provider.createModelStructureQueryConnectorInstance();
	}

	private static ConnectorProvider getConnectorProvider(String family,
			Class<? extends QueryConnector> connectorClass) {
		BundleContext context = Activator.getContext();
		ServiceReference<ConnectorRegistry> registryRef = context.getServiceReference(ConnectorRegistry.class);
		ConnectorRegistry registry = context.getService(registryRef);
		ServiceReference<ConnectorProvider> providerRef = registry.requestServiceReference(family, connectorClass);
		return context.getService(providerRef);
	}

	/*
	 * Existing Method reused. Have a look on
	 * PerformanceMetricsQueryInterpreterImpl
	 */
	public static void processSelectClause(final WhatIfQuery selectQuery, final Map<String, Entity> aliasMap) {
		final MetricReferenceClauses metricRefClauses = selectQuery.getMetricReferenceClauses();
		final EList<MetricClauseType> metricClauseTypes = metricRefClauses.getMetricClauseTypes();

		EntityMapping mapping = initializeEntityMapping(selectQuery);
		for (final MetricClauseType ref : metricClauseTypes) {
			final Probe probe = domainFactory.createProbe();
			Entity entity = null;

			if (ref instanceof MetricClause) {
				// Found a plain metric
				final MetricClause metricClause = (MetricClause) ref;

				final MetricReference metricRef = metricClause.getMetricReference();
				final String entityRef = metricRef.getEntityReference();
				final String metricRef2 = metricRef.getMetricReference();
				String statTypeRef = null;
				if (metricRef.getStatTypeReference() != null)
					statTypeRef = metricRef.getStatTypeReference();
				else
					statTypeRef = getDefaultStatType(entityRef, metricRef2, mapping);

				entity = aliasMap.get(entityRef);
				probe.setMetricName(metricRef2);
				final StatType st = domainFactory.createStatType();
				st.setTypeName(statTypeRef);
				probe.getStatTypes().add(st);
			}
			final EList<Probe> probes = entity.getProbes();
			probes.add(probe);
		}
	}

	private static String getDefaultStatType(String entityRef, String metricRef2, EntityMapping mapping) {
		for (Entity e : mapping.getEntities())
			if (e.getAlias().equals(entityRef))
				for (Probe p : e.getProbes())
					if (p.getMetricName().equals(metricRef2))
						if (p.getStatTypes().isEmpty())
							return "DEFAULT";
						else
							return p.getStatTypes().get(0).getTypeName();
		throw new IllegalStateException("There is no Metric " + entityRef + "." + metricRef2);
	}

	private static EntityMapping initializeEntityMapping(WhatIfQuery selectQuery) {
		UsingClause uc = selectQuery.getUsingClause();
		ForClause fc = selectQuery.getForClause();

		final DescartesQL metricsQuery = dqlFactory.createDescartesQL();
		final ListQuery listQuery = dqlFactory.createListQuery();
		metricsQuery.setQuery(listQuery);
		listQuery.setUsingClause(uc);
		final ListMetricsQuery listMetricsQuery = dqlFactory.createListMetricsQuery();
		listQuery.setQueryType(listMetricsQuery);
		listMetricsQuery.setMode("LIST METRICS");
		listMetricsQuery.setEntityReferenceClause(fc.getEntityReferenceClause());

		final QueryExecutionEngine qee = new QueryExecutionEngineImpl();
		return qee.execute(metricsQuery).get(0);
	}

	/*
	 * Existing Method reused. Have a look on
	 * PerformanceMetricsQueryInterpreterImpl
	 */
	public static void processForClause(final WhatIfQuery WhatIfQuery, final Map<String, Entity> aliasMap,
			final EntityMapping request) {
		final List<EntityReference> entityRefs = WhatIfQuery.getForClause().getEntityReferenceClause()
				.getEntityReferences();

		for (final EntityReference ref : entityRefs) {
			Entity entity = domainFactory.createEntity();
			entity.setEntityType(ref.getType());
			final AliasClause aliasClause = ref.getAlias();

			if (aliasClause != null) {
				entity.setAlias(aliasClause.getAlias());
			}

			entity.setIdentifier(ref.getIdentifier());

			if (aliasMap.put(entity.getAlias(), entity) != null) {
				throw new IllegalArgumentException("Duplicate aliases in FOR, aborting.");
			}
			request.getEntities().add(entity);
		}
	}

	/*
	 * generating List of values, starting with a state value
	 */
	public static HashMap<String, List<String>> processWhatIfVariation(final WhatIfQuery whatIfQuery,
			final HashMap<String, String> hashMap) {

		LinkedList<WhatIfVariationClause> numberOfVariables = new LinkedList<WhatIfVariationClause>(
				whatIfQuery.getWhatIfVariation().getWhatIfVariationClause());

		HashMap<String, List<String>> results = new HashMap<String, List<String>>();

		for (int i = 0; i < numberOfVariables.size(); i++) {

			EList<EObject> variations = numberOfVariables.get(i).getValues().getVariationClause();

			LinkedList<String> newValues = new LinkedList<String>();
			newValues.add(hashMap.get(numberOfVariables.get(i).getVariationPoint()).toString());

			for (EObject variation : variations) {

				// Variation with IntervalVariationClause (ex.: "10 .. 20 BY 5")
				if (variation instanceof IntervalVariationClause) {
					IntervalVariationClause clause = (IntervalVariationClause) variation;
					List<String> values = new LinkedList<String>();
					int current = clause.getStart();
					int stop = clause.getEnd();
					int step = clause.getStep();
					while (current <= stop) {
						values.add(String.valueOf(current));
						current = current + step;
					}
					if (Integer.valueOf(values.get(values.size() - 1)) < stop) {
						values.add(String.valueOf(stop));
					}
					for (int k = 0; k < values.size(); k++) {
						if (newValues.contains(values.get(k))) {
							logger.info("Integer " + values.get(k) + " already exists and was removed!");
							values.remove(k);
						}
					}
					newValues.addAll(values);

					// Variation with ValueVariationClause (ex.: "10, 20, 30")
				} else if (variation instanceof ValueVariationClause) {
					ValueVariationClause clause = (ValueVariationClause) variation;
					List<Integer> values = clause.getValue();
					for (Integer integer : values) {
						if (!newValues.contains(integer.toString())) {
							newValues.add(integer.toString());
						} else
							logger.info("Integer " + integer + " already exists and was removed!");
					}

					// Variation with ArithmeticClauses (ex.: "ADD 10, MULTIPLY
					// 0.5")
				} else if (variation instanceof ArithmeticClause) {

					ArithmeticClause clause = (ArithmeticClause) variation;
					int newValue = 0;
					if (clause.getMode().equals("ADD")) {
						if (clause.getValue().getIntValue() == 0) {
							newValue = (int) (Integer.valueOf(hashMap.get(numberOfVariables.get(i).getVariationPoint()))
									+ clause.getValue().getDoubleValue());
							logger.info("Attention: Int-Cast");

						} else {
							newValue = Integer.valueOf(hashMap.get(numberOfVariables.get(i).getVariationPoint()))
									+ clause.getValue().getIntValue();

						}

					} else if (clause.getMode().equals("SUB")) {
						if (clause.getValue().getIntValue() == 0) {
							newValue = (int) (Integer.valueOf(hashMap.get(numberOfVariables.get(i).getVariationPoint()))
									- clause.getValue().getDoubleValue());
							logger.info("Attention: Int-Cast");
							if (newValue < 0)
								throw new IllegalArgumentException("Illegal Argument for variationPoint "
										+ numberOfVariables.get(i).getVariationPoint() + " with subtraction of value "
										+ clause.getValue().getDoubleValue() + " with result of " + newValue);

						} else {
							newValue = Integer.valueOf(hashMap.get(numberOfVariables.get(i).getVariationPoint()))
									- clause.getValue().getIntValue();
							if (newValue < 0)
								throw new IllegalArgumentException("Illegal Argument for variationPoint "
										+ numberOfVariables.get(i).getVariationPoint() + " with subtraction of value "
										+ clause.getValue().getIntValue() + " with result of " + newValue);

						}

					} else if (clause.getMode().equals("MULTIPLY")) {
						if (clause.getValue().getIntValue() == 0) {
							newValue = (int) (Integer.valueOf(hashMap.get(numberOfVariables.get(i).getVariationPoint()))
									* clause.getValue().getDoubleValue());
							logger.info("Attention: Int-Cast");

						} else {
							newValue = Integer.valueOf(hashMap.get(numberOfVariables.get(i).getVariationPoint()))
									* clause.getValue().getIntValue();

						}

					} else
						throw new IllegalArgumentException(
								"Invalid arithemtic instruction for variationPoint " + numberOfVariables.get(i));

					if (!newValues.contains(newValue)) {
						newValues.add(String.valueOf(newValue));
					} else
						logger.info("Integer " + newValue + " already exists and was removed!");
				} else if (variation instanceof VariationInstruction) {
					VariationInstruction clause = (VariationInstruction) variation;
					String newValue = clause.getVariation() + ";;;;;" + clause.getArgument();
					newValues.add(newValue);

				} else
					throw new IllegalArgumentException(
							"Invalid Variation for variationPoint " + numberOfVariables.get(i));

			}

			results.put(numberOfVariables.get(i).getVariationPoint(), newValues);

		}

		return results;
	}

	/*
	 * mainly copied from DoubleFilter applyFilter()
	 */
	public static EntityMapping createDeltas(EntityMapping initial, EntityMapping modified) {
		if (!initial.equals(modified)) { // else the initial gets modified and
			// looses all content
			// checking all Resouces
			for (Iterator<Entity> resourceIterator = modified.getEntities().iterator(); resourceIterator.hasNext();) {
				Entity resource = resourceIterator.next();
				for (Iterator<Probe> probeIterator = resource.getProbes().iterator(); probeIterator.hasNext();) {
					Probe probe = probeIterator.next();
					for (Iterator<StatType> stIterator = probe.getStatTypes().iterator(); stIterator.hasNext();) {
						StatType st = stIterator.next();
						Result r = st.getResult();
						if (r instanceof DecimalResult) {
							DecimalResult modifiedResult = (DecimalResult) r;
							DecimalResult initialResult = getResult(initial, resource.getIdentifier(),
									probe.getMetricName(), st.getTypeName());

							modifiedResult.setValue(modifiedResult.getValue().subtract(initialResult.getValue()));

						}
					}
				}
			}

			return modified;
		} else
			return initial;

	}

	/*
	 * copied from DoubleFilter
	 */

	private static DecimalResult getResult(final EntityMapping initial, String identifier, String metricIdentifier,
			String string) {
		for (Iterator<Entity> resourceIterator = initial.getEntities().iterator(); resourceIterator.hasNext();) {
			Entity resource = resourceIterator.next();
			if (resource.getIdentifier().equals(identifier)) {
				for (Iterator<Probe> probeIterator = resource.getProbes().iterator(); probeIterator.hasNext();) {
					Probe probe = probeIterator.next();
					if (probe.getMetricName().equals(metricIdentifier)) {
						for (StatType st : probe.getStatTypes())
							if (st.getTypeName().equals(string))
								return (DecimalResult) st.getResult();
					}
				}
			}
		}
		return null;
	}

	/*
	 * creates an empty EntityMapping with model location
	 */
	public static EntityMapping createEmptyEntityMapping(String modelLocation) {
		EntityMapping mapping = mappingFactory.createEntityMapping();
		mapping.setModelLocation(modelLocation);
		return mapping;
	}
}
