/**
 * ==============================================
 *  DQL Adapter for SimQPN
 * ==============================================
 *
 * (c) Copyright 2014-2016, by Simon Eismann, Juergen Walter and other contributors.
 *
 * All rights reserved. This software is made available under the terms of the
 * Eclipse Public License (EPL) v1.0 as published by the Eclipse Foundation
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This software is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the Eclipse Public License (EPL)
 * for more details.
 *
 * You should have received a copy of the Eclipse Public License (EPL)
 * along with this software; if not visit http://www.eclipse.org or write to
 * Eclipse Foundation, Inc., 308 SW First Avenue, Suite 110, Portland, 97204 USA
 * Email: license (at) eclipse.org
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */
package tools.descartes.dql.whatif.tests.util;

import java.io.File;
import java.io.IOException;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.LinkedList;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.log4j.Logger;
import org.dom4j.DocumentException;
import org.dom4j.Element;
import org.dom4j.io.SAXReader;
import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.IPath;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;

import tools.descartes.dql.core.engine.util.DQLLogger;

public class FileUtil {
	private static final Logger log = DQLLogger.getLogger(FileUtil.class.getName());
	
	public static String getPath(String name) {
		URL url = FileUtil.class.getClassLoader().getResource(name);
		if (url.toString().contains("bundleresource:"))
			try {
				url = FileLocator.resolve(url);
			} catch (IOException e) {
				log.error("couldn't resolve the URL:" + url.toString());
			}
		File modelPath = null;
		try {
				modelPath = new File(url.toURI());
			} catch (URISyntaxException e) {
				log.error("UriSyntaxException " + e.getMessage());
				throw new IllegalStateException("A internal error occured while trying to get the"
						+ "path of a file!");
			}
		return getURL(modelPath);
	}
	
	public static Document loadDocument(File document) {
		try {
			DocumentBuilderFactory dbFactory = DocumentBuilderFactory.newInstance();
			DocumentBuilder dBuilder = null;
				dBuilder = dbFactory.newDocumentBuilder();
			Document doc = null;
				doc = dBuilder.parse(document);
			doc.getDocumentElement().normalize();
			return doc;
		} catch (SAXException | IOException | ParserConfigurationException e) {
			log.error("Couldn't load document");
			throw new IllegalStateException("Couldn't load document");
		}
	}
	
	public static String getURL(File f) {
		return f.getAbsolutePath();
	}
	public static File createFile(String filePath) {
		File f = new File(filePath);
		try {
			f.createNewFile();
		} catch (IOException e) {
			log.error("Was unable to create required Folders/Files!");
			log.error(e.getMessage());
		}
		return f;
	}
	
	public static String getNameOfFile(String filePath) {
		return filePath.substring(filePath.lastIndexOf(File.separator) + 1);
	}
	
	public static String getModelPath(String modelLocation) {
		if (isLocalPath(modelLocation)) {
			modelLocation = parseLocalPath(modelLocation);
		}
		return modelLocation;
	}
	
	public static Element loadXmlFromFile(String modelLocation) {
		modelLocation = getModelPath(modelLocation);
		
		SAXReader xmlReader = new SAXReader();
		org.dom4j.Document inputDoc = null;
		try {
			inputDoc = xmlReader.read(new File(modelLocation));
		} catch (DocumentException e2) {
			log.error(e2.getMessage());
			throw new IllegalStateException("Could not find the specified file or it is not a QPN net!"
					+ " \n\tPath:\n\t" + modelLocation);
		}
		return inputDoc.getRootElement();
	}
	
	private static String parseLocalPath(String path) {
		IPath workspaceLocation = ResourcesPlugin.getWorkspace()
				.getRoot().getLocation();
		IPath workspaceModelLocation = workspaceLocation
				.append(path);
		File file = workspaceModelLocation.toFile();
		if (!file.exists())
			file = findFileInWorkspace(path);
		path = FileUtil.getURL(file);
		return path;
	}
	
	private static File findFileInWorkspace(String name) {
		IWorkspaceRoot root = ResourcesPlugin.getWorkspace().getRoot();
		List<File> foundFiles = getFilesWithNameInside(root, name);
		if (foundFiles.size() == 0)
			throw new IllegalStateException("Couldn't find File " + name);
		if (foundFiles.size() > 1)
			throw new IllegalStateException("The Filename " + name + " is not unique");
		return foundFiles.get(0);
	}

	private static List<File> getFilesWithNameInside(IContainer container,
			String name) {
		List<File> files = new LinkedList<File>();
		for (IResource res: getChildren(container)) {
			if (res instanceof IFile) {
				IFile f = (IFile) res;
				if (getName(f).equals(name)) {
					addToList(f, files);
				}
			}
			if (res instanceof IContainer) {
				IContainer con = (IContainer) res; 
				recursiveCall(con, name, files);
			}
		}
		return files;
	}
	
	private static void recursiveCall(IContainer con, String name, List<File> files) {
		if (!con.isHidden() && con.isAccessible() && con.exists() && !con.isDerived())
			files.addAll(getFilesWithNameInside(con, name));
	}

	private static void addToList(IFile f, List<File> files) {
		files.add(new File(f.getLocationURI()));
	}

	private static String getName(IFile f) {
		return f.getName();
	}

	private static IResource[] getChildren(IContainer container) {
		try {
			return container.members();
		} catch (CoreException e) {
			return new IResource[0];
		}
	}

	private static boolean isLocalPath(String path) {
		//check for data because of jenkins path^^
		return (path.matches("(\\\\|\\/).*") || !path.contains(":")) 
				&& !path.matches("(\\\\home|\\/home).*")
				&& !path.matches("(\\\\data|\\/data).*");
	}
}
