/**
 * ==============================================
 *  DQL : Descartes Query Language
 * ==============================================
 *
 * (c) Copyright 2014-2016, by Juergen Walter, Fabian Gorsler, Fabian Brosig and other contributors.
 *
 * Project Info:   http://descartes.tools/dql
 *
 * All rights reserved. This software is made available under the terms of the
 * Eclipse Public License (EPL) v1.0 as published by the Eclipse Foundation
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This software is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the Eclipse Public License (EPL)
 * for more details.
 *
 * You should have received a copy of the Eclipse Public License (EPL)
 * along with this software; if not visit http://www.eclipse.org or write to
 * Eclipse Foundation, Inc., 308 SW First Avenue, Suite 110, Portland, 97204 USA
 * Email: license (at) eclipse.org
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package tools.descartes.pavo.export.pdf;

import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;

import org.jfree.chart.JFreeChart;

import com.itextpdf.text.Document;
import com.itextpdf.text.DocumentException;
import com.itextpdf.text.Font;
import com.itextpdf.text.FontFactory;
import com.itextpdf.text.Image;
import com.itextpdf.text.Paragraph;
import com.itextpdf.text.Rectangle;
import com.itextpdf.text.pdf.BaseFont;
import com.itextpdf.text.pdf.PdfWriter;

import tools.descartes.pavo.controller.IExportController;
import tools.descartes.pavo.result.ResultContainerPavo;
import tools.descartes.pavo.view.ExportDialogFrame;

/**
 * Export controller extension using the PAVO ExtensionPoint. The exported file
 * is of type PDF with the width and height taken from the user interaction with
 * the GUI Uses the IText library
 * 
 * @see tools.descartes.pavo.export.pdf.itext
 * @author Maximilian König <max.c.koenig@t-online.de>
 * @since 2016
 */

public class PDFExportController implements IExportController {
	String filename;

	@Override
	public void export(File file, JFreeChart chart, ResultContainerPavo result) {
		filename = this.updatePathToSuffix(file).getPath();
		try {
			this.create(new FileOutputStream(this.updatePathToSuffix(file)), chart, ExportDialogFrame.widthStatic,
					ExportDialogFrame.heightStatic);
		} catch (DocumentException | IOException e) {
			e.printStackTrace();
		}

	}

	@Override
	public String getStringForCombo() {
		return new String("PDF Export");
	}

	@Override
	public String getIdentifier() {
		return new String("PDF Export");
	}

	/**
	 * Creates PDf file.
	 * 
	 * @param outputStream
	 *            {@link OutputStream}.
	 * @throws DocumentException
	 * @throws IOException
	 */
	public void create(OutputStream outputStream, JFreeChart chart, int width, int height)
			throws DocumentException, IOException {
		Document document = null;
		PdfWriter writer = null;

		try {

			BufferedImage bufferedImage = chart.createBufferedImage(width, height);
			Rectangle pagesize = new Rectangle(bufferedImage.getWidth(), bufferedImage.getHeight() + 10);

			document = new Document(pagesize);

			writer = PdfWriter.getInstance(document, outputStream);
			document.open();
			FontFactory.defaultEmbedding = true;

			BaseFont bf1 = BaseFont.createFont(chart.getXYPlot().getRangeAxis().getLabelFont().getName(),
					BaseFont.WINANSI, true);
			bf1.setSubset(true);
			Font font1 = new Font(bf1, 1);
			document.add(new Paragraph(" ", font1));
			//
			BaseFont bf2 = BaseFont.createFont(chart.getXYPlot().getRangeAxis().getLabelFont().getName(),
					BaseFont.WINANSI, true);
			bf2.setSubset(true);
			Font font2 = new Font(bf2, 1);
			document.add(new Paragraph(" ", font2));

			Image image = Image.getInstance(writer, bufferedImage, 1.0f);
			image.setAbsolutePosition(0, 0);
			document.add(image);
			// release resources
			document.close();
			document = null;

			writer.close();
			writer = null;
		} catch (DocumentException de) {
			throw de;
		} catch (IOException ioe) {
			throw ioe;
		} finally {
			// release resources
			if (null != document) {
				try {
					document.close();
				} catch (Exception ex) {
				}
			}

			if (null != writer) {
				try {
					writer.close();
				} catch (Exception ex) {
				}
			}
		}
	}

	@Override
	public File updatePathToSuffix(File file) {
		String name = file.getPath();
		if (name.length() < 3 || !name.substring(name.length() - 3, name.length()).equals("pdf")) {
			return new File(name + ".pdf");
		} else {
			return file;
		}
	}

}